#include <zephyr.h>
#include <device.h>
#include <drivers/gpio.h>
#include <sys/__assert.h>
#include <string.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/conn.h>
#include <bluetooth/gatt.h>
#include <bluetooth/mesh/models.h>
#include <bluetooth/mesh/dk_prov.h>
#include <dk_buttons_and_leds.h>

#include "model_handler.h"

#define LOG_LEVEL LOG_LEVEL_DBG
#include <logging/log.h>
LOG_MODULE_REGISTER(my_ble_sample);

#define DEVICE_NAME             CONFIG_BT_DEVICE_NAME
#define DEVICE_NAME_LEN         (sizeof(DEVICE_NAME) - 1)


static const struct bt_data ad[] = {
    BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
    BT_DATA(BT_DATA_NAME_COMPLETE, DEVICE_NAME, DEVICE_NAME_LEN),
};


typedef enum
{
    MODE_BLE_MESH     = 0x00,
    MODE_NOT_BLE_MESH = 0x01,
    NUMBER_OF_BLE_MODES
} ble_mode_t;

static struct k_delayed_work change_ble_mode_work;

static void button_pressed(uint32_t button_state, uint32_t has_changed);
static void change_ble_mode(struct k_work *work);
static void connected(struct bt_conn *conn, uint8_t err);
static void disconnected(struct bt_conn *conn, uint8_t reason);
static void start_ble(void);
static void stop_ble(void);
static void start_ble_mesh(void);
static void stop_ble_mesh();
static void bt_ready(int err);


static void button_pressed(uint32_t button_state, uint32_t has_changed)
{
    if (has_changed & DK_BTN1_MSK) {  /* Check a particular button whether its state has changed */
        if ((button_state & DK_BTN1_MSK) != 0) {  /* The button is pressed */
            k_delayed_work_submit(&change_ble_mode_work, K_NO_WAIT);
            return;
        }
    }
}


static void change_ble_mode(struct k_work *work)
{
    static ble_mode_t current_ble_mode = MODE_BLE_MESH;

    switch(current_ble_mode) {
        case MODE_BLE_MESH:
            LOG_DBG("Change mode from MODE_BLE_MESH to MODE_NOT_BLE_MESH.");
            stop_ble_mesh();
            start_ble();
            break;
        case MODE_NOT_BLE_MESH:
            LOG_DBG("Change mode from MODE_NOT_BLE_MESH to MODE_BLE_MESH.");
            stop_ble();
            start_ble_mesh();
            break;
    }

    current_ble_mode = (current_ble_mode + 1) % NUMBER_OF_BLE_MODES;
}


static void connected(struct bt_conn *conn, uint8_t err)
{
    if (err) {
        LOG_ERR("Connection failed (err 0x%02x)", err);
    } else {
        LOG_INF("Connected");
    }
}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
    LOG_INF("Disconnected (reason 0x%02x)", reason);
    start_ble();
}

static struct bt_conn_cb conn_callbacks = {
    .connected = connected,
    .disconnected = disconnected,
};

static void start_ble(void)
{
    int err;

    err = bt_le_adv_stop();
    if (err) {
        LOG_ERR("Failed to stop advertising (err %d)", err);
    } else {
        LOG_INF("Stop advertising");
    }

    err = bt_le_adv_start(BT_LE_ADV_CONN_NAME, ad, ARRAY_SIZE(ad), NULL, 0);
    if (err) {
        LOG_ERR("Failed to start advertising (err %d)", err);
        return;
    } else {
        LOG_INF("Start advertising");
    }

    bt_conn_cb_register(&conn_callbacks);
}


static void stop_ble(void)
{
    int err;

    bt_conn_cb_register(NULL);

    err = bt_le_adv_stop();
    if (err) {
        LOG_ERR("Failed to stop advertising (err %d)", err);
    } else {
        LOG_INF("Stop advertising");
    }
}


static void start_ble_mesh(void)
{
    int err;

    if (bt_mesh_is_provisioned()) {
        err = bt_mesh_resume();
        if (err) {
            LOG_ERR("Failed to resume BLE mesh (err %d)", err);
        } else {
            LOG_INF("Resume BLE mesh");
        }
    } else {
        err = bt_mesh_prov_enable(BT_MESH_PROV_ADV | BT_MESH_PROV_GATT);
        if (err) {
            LOG_ERR("Failed to enable BLE mesh provisioning (err %d)", err);
        } else {
            LOG_INF("Enable BLE mesh provisioning");
        }
    }
}


static void stop_ble_mesh()
{
    int err;
    
    if (bt_mesh_is_provisioned()) {
        err = bt_mesh_suspend();
        if (err) {
            LOG_ERR("Failed to suspend BLE mesh (err %d)", err);
        } else {
            LOG_INF("Suspend BLE mesh");
        }
    } else {
        err = bt_mesh_prov_disable(BT_MESH_PROV_ADV | BT_MESH_PROV_GATT);
        if (err) {
            LOG_ERR("Failed to disable BLE mesh provisioning (err %d)", err);
        } else {
            LOG_INF("Disable BLE mesh provisioning");
        }
    }    
}


static void bt_ready(int err)
{
    if (err) {
        LOG_ERR("Bluetooth init failed (err %d).", err);
        return;
    } else {
        LOG_INF("Bluetooth initialized.");
    }

    err = bt_mesh_init(bt_mesh_dk_prov_init(), model_handler_init());
    if (err) {
        LOG_ERR("Initializing mesh failed (err %d)\n", err);
        return;
    } else {
        LOG_INF("Initialize BLE mesh");
    }

    if (IS_ENABLED(CONFIG_SETTINGS)) {
        settings_load();
    }

    /* This will be a no-op if settings_load() loaded provisioning info */
    err = bt_mesh_prov_enable(BT_MESH_PROV_ADV | BT_MESH_PROV_GATT);
    if (err) {
        LOG_ERR("Failed to enable BLE mesh provisioning (err %d)", err);
    } else {
        LOG_INF("Enable BLE mesh provisioning");
    }

    k_delayed_work_init(&change_ble_mode_work, change_ble_mode);
}


void main(void)
{
    int err;

    err = dk_leds_init();
    if (err) {
        LOG_ERR("LEDs init error %d", err);
        return;
    }

    err = dk_buttons_init(button_pressed);
    if (err) {
        LOG_ERR("Buttons init error %d", err);
        return;
    }

    err = bt_enable(bt_ready);
    if (err) {
        LOG_ERR("Bluetooth init failed (err %d).", err);
        return;
    }
}
